!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Original module written by Julien Lesgourgues for using weak lensing data in CosmoMC        !
! It was written for the COSMOS 3D data by Massey et al. [astro-ph/0701480],                  !
! was used in [arXiv:0705.0533] by J. Lesgourgues, M. Viel, M.G. Haehnelt, R. Massey          !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This modified version was written by Jason Dossett.                                         !
! It allows the use to the CFHTLens weak lensing tomography data of [arXiv:1303.1808]         !
! and was first used in [arXiv:1501.03119] by J. Dossett, M. Ishak, D. Parkinson, and T. Davis!
! to probe deviations from General Relativity.                                                !
! Last Modified 01/2015                                                                       !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

MODULE CFHTLens
  use Camb, only : f_k_real
  use constants, ONLY: const_pi
  use cmbtypes
  use likelihood
  use powerspec
  use Angular ! ancillary module converting harmonic power spectrum into angular correlation functions
  implicit none
  ! precision parameters:
  integer, parameter :: nlmax=130  ! fixes the maximum required l value in C_l(lensing) 
  real, parameter :: dlnl=0.1      ! fixes logarithmic l step for C_l(lensing)
  real, parameter :: xstop=200.    ! fixes maximum value of x =  l * theta in angular correlation function 
  ! note: integration of C_l J_0,4(l*theta) over l will stop 
  ! at l = min( ll(nlmax),  xstop/theta )

  ! parameters describing CFHTLens data set:

  integer, parameter :: nbinmax=6  ! Number of redshift bins

  integer, parameter :: nzmax = 72 ! Number of z values in galaxy distribution file

  integer, parameter :: nthetamax=5 ! number of angles theta

  integer, parameter :: ncl = nbinmax*(nbinmax+1)/2    ! number of Cross-power spectra to compute

!Data Arrays
  real z(nzmax),etaz_f(nzmax,nbinmax),etaz_b(nzmax,nbinmax),eta_norm_f(nbinmax),eta_norm_b(nbinmax) ! data's selection functions
!JD Finished to here  
  real theta(nthetamax) ! data array of theta values
  !Arrays for calculating -ln(Like)
  real d_theo(2*nthetamax*ncl)   ! Theoretical shear cross-correlations, data arranged like a vector
  real d_obs(2*nthetamax*ncl)  ! Observed shear cross-correlations, data arranged like a vector
  real(mcp) invcov(2*nthetamax*ncl,2*nthetamax*ncl)  !inverse Covariance Matrix

  !Allocatable Arrays
  !Dimension Theory%num_k-1
  real, allocatable :: kk(:)
  !Dimension num_power_redshifts
  real, allocatable :: PP_GG(:), ddPP_GG(:)   ! P_GG(k=l/r, z=z(r))
  real, allocatable :: PP_GI(:), ddPP_GI(:)   ! P_GI(k=l/r, z=z(r))
  real, allocatable :: PP_II(:), ddPP_II(:)   ! P_GI(k=l/r, z=z(r))
  !Dimension Theory%num_k-1
  real, allocatable :: u_spline(:), u_spline_GG(:), u_spline_GI(:), u_spline_II(:)
  !Dimension num_power_redshifts
  real, allocatable :: zz(:),rr(:),dz_over_dr(:)
  real, allocatable :: ddrr(:),dddz_over_dr(:)

  type, extends(CosmologyLikelihood) :: CFHTLensLikelihood
    contains
    procedure :: LogLike => CFHTLensLnLike
    procedure :: ReadIni => CFHTLensReadIni
  end type CFHTLensLikelihood


CONTAINS

  subroutine CFHTLensLikelihood_Add(LikeList, Ini)
    use IniFile
    use settings
    class(LikelihoodList) :: LikeList
    Type(TIniFile) :: ini
    Type(CFHTLensLikelihood), pointer :: like

    use_CFHTLS = Ini_Read_Logical_File(Ini, 'use_CFHTLS',.false.)
    if (.not. use_CFHTLS) return
    
    allocate(like)
    Like%LikelihoodType = 'WeakLen'
    like%needs_powerspectra = .true.
    like%needs_nonlinear_pk = .true.
    like%max_z = 3
    like%num_z = 20
    like%kmax = 100
    call Like%loadParamNames(trim(DataDir)//'CFHTLenS/CFHTLenS.paramnames')
    call like%ReadDatasetFile(ReadIniFileName(Ini,'CFHTLens_dataset'))
    call LikeList%Add(like)

  end subroutine CFHTLensLikelihood_Add

  subroutine CFHTLensReadIni(like,Ini)
    use MatrixUtils
    class(CFHTLensLikelihood) like
    Type(TIniFile) :: Ini
    real d_read(2*nthetamax,nbinmax*(nbinmax+1)/2)  !temp for correlation vector
    real theta_dum
    integer nz, k, ios
    character(LEN=Ini_max_string_len) prename,name,number

      if(feedback>0)write(*,*)'Initializing refined CFHTLens weak lensing data'
      !Read galaxy distribution files
      prename = ReadIniFileName(Ini,'redshift_dist_prefix_foreground')
      do k=1,nbinmax
        write(number,'(I0)')k
        name=trim(prename)//'_'//trim(number)//'.hist'
        open (unit=10,file=name)
        nz=1
        do while (nz<=nzmax)
          read(10,*,IOSTAT=ios) z(nz), etaz_f(nz,k)
          if(IS_IOSTAT_END(ios)) Exit
          if(ios.ne.0) cycle
          nz=nz+1
        end do
      end do
      close(10)

      prename = ReadIniFileName(Ini,'redshift_dist_prefix_background',NotFoundFail = .false.)
      if(prename =='') then 
        etaz_b = etaz_f
      else
        do k=1,nbinmax
          write(number,'(I0)')k
          name=trim(prename)//'_'//trim(number)//'.hist'
          open (unit=10,file=name)
          nz=1
          do while (nz<=nzmax)
            read(10,*,IOSTAT=ios) z(nz), etaz_b(nz,k)
            if(IS_IOSTAT_END(ios)) Exit
            if(ios.ne.0) cycle
            nz=nz+1
          end do
        end do
        close(10)
      end if

      !read data vector
      name=ReadIniFileName(Ini,'measurements_file')
      open (unit=10,file=name)
      do k=1,2*nthetamax
        if(k .le. nthetamax) then
          read(10,*)theta(k),d_read(k,:)
          theta(k)=theta(k)/60.*const_pi/180.
        else
          read(10,*)theta_dum,d_read(k,:)
        end if
      end do
      close(10)

      !rearrange datavector to match covariance matrix
      do k=1,2*nthetamax
        do nz=1,ncl
          d_obs(k+2*nthetamax*(nz-1))=d_read(k,nz)
        end do
      end do

      ! Read covariance matrix
      name=ReadIniFileName(Ini,'covmat_file')
      open (unit=10,file=name)
      do k=1,2*nthetamax*ncl
          read(10,*)invcov(k,:)
      end do
      close(10)
      !get inverse covariance matrix
      call Matrix_Inverse(invcov)
      !Correction factor as seen in Heymans et al arXiv:1303.1808
      !Section 3.3.1.  n_\mu=1656 and p=210
      invcov = (1656.-210.-2.)/(1656.-1.)*invcov

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! integrate eta(z) over z (in view of normalizing it to one) !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      eta_norm_f(:)=0.
      eta_norm_b(:)=0.
      do nz=2,nzmax
          eta_norm_f(:)=eta_norm_f(:)+0.5*(etaz_f(nz-1,:)+etaz_f(nz,:))*(z(nz)-z(nz-1))
          eta_norm_b(:)=eta_norm_b(:)+0.5*(etaz_b(nz-1,:)+etaz_b(nz,:))*(z(nz)-z(nz-1))
      end do  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Compute Bessel interpolation array !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      call GetBessels(xstop*1.d0) ! subroutine of the "Angular" module

  end subroutine CFHTLensReadIni
      
  subroutine InitLens(Theory)
  Class(TheoryPredictions) Theory
    if(allocated(PP_GG))deallocate(PP_GG)
    if(allocated(ddPP_GG))deallocate(ddPP_GG)
    if(allocated(PP_GI))deallocate(PP_GI)
    if(allocated(ddPP_GI))deallocate(ddPP_GI)
    if(allocated(PP_II))deallocate(PP_II)
    if(allocated(ddPP_II))deallocate(ddPP_II)
    if(allocated(u_spline))deallocate(u_spline)
    if(allocated(u_spline_GG))deallocate(u_spline_GG)
    if(allocated(u_spline_GI))deallocate(u_spline_GI)
    if(allocated(u_spline_II))deallocate(u_spline_II)
    if(allocated(zz))deallocate(zz)
    if(allocated(rr))deallocate(rr)
    if(allocated(dz_over_dr))deallocate(dz_over_dr)
    if(allocated(ddrr))deallocate(ddrr)
    if(allocated(dddz_over_dr))deallocate(dddz_over_dr)
    if(allocated(kk))deallocate(kk)
    allocate(PP_GG(num_power_redshifts))
    allocate(ddPP_GG(num_power_redshifts))
    allocate(PP_GI(num_power_redshifts))
    allocate(ddPP_GI(num_power_redshifts))
    allocate(PP_II(num_power_redshifts))
    allocate(ddPP_II(num_power_redshifts))
    allocate(zz(num_power_redshifts))
    allocate(rr(num_power_redshifts))
    allocate(dz_over_dr(num_power_redshifts))
    allocate(ddrr(num_power_redshifts))
    allocate(dddz_over_dr(num_power_redshifts))
    allocate(kk(Theory%num_k-1))
    allocate(u_spline(Theory%num_k-1))
    allocate(u_spline_GG(Theory%num_k-1))
    allocate(u_spline_GI(Theory%num_k-1))
    allocate(u_spline_II(Theory%num_k-1))
  end subroutine InitLens
  
  function CFHTLensLnLike(like,CMB,Theory,DataParams)
    Class(CFHTLensLikelihood) :: like
    Class(CMBParams) CMB
    Class(TheoryPredictions) Theory
    real(mcp) DataParams(:)
    real(mcp) :: CFHTLensLnLike
    
    integer nl,nc,clc1,clc2
    real ll(nlmax)
    integer nkmax,nk
    integer zix

    
    ! for spline:
    real sig_spline,p_spline
    real p_spline_GG, p_spline_GI, p_spline_II
    ! for splint:
    integer k,khi,klo
    real a,b,h
    ! for eta(z):
    integer nz,nbin,nr,nrmax,nrs
    real r(nzmax),dz_o_dr(nzmax),etar_f(nzmax,nbinmax),etar_b(nzmax,nbinmax)
    real g_b(nzmax,nbinmax),g_f(nzmax,nbinmax)
    ! for Theory%z_and_r(:,zix)
    ! For C_l
    real(mcp) kvalue
    real PP_GG_r, PP_GI_r,PP_II_r
    real Cl_integrand_GG(nzmax,ncl)
    real Cl_integrand_GI(nzmax,ncl)
    real Cl_integrand_II(nzmax,ncl)
    real C_GG(nlmax,ncl),ddC_GG(nlmax,ncl)
    real C_GI(nlmax,ncl),ddC_GI(nlmax,ncl)
    real C_II(nlmax,ncl),ddC_II(nlmax,ncl)
    !For C(theta)
    real Cp(nthetamax,ncl),Cm(nthetamax,ncl) !cross-correlation arrays for function Cl to Ctheta
    real Cp_GG(nthetamax,ncl),Cm_GG(nthetamax,ncl)
    real Cp_GI(nthetamax,ncl),Cm_GI(nthetamax,ncl)
    real Cp_II(nthetamax,ncl),Cm_II(nthetamax,ncl)
    real CFHTLens_A


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!! Call function to allocate arrays !!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    call InitLens(Theory)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!! Various steps for Likelihood computation start here !!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! JD: A_CFHTLens for intrinsic alignment correction  !!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    CFHTLens_A=DataParams(1)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Convert eta(z) to eta(r) !!!!!!!!!!!!!!!!!!!!
!!! (first unnormalized, second normalized to one) !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    zz(:)=Theory%redshifts(:)  ! note that z (and r, ...) must be rearranged in growing order for splint
    ! read array Theory%z_and_r(2,zix) = list of corresponding comoving distances rr(zix)
    ! read array Theory%z_and_r(3,zix) = list of corresponding dz/dr (= Hubble parameter)    
    rr(:)=Theory%z_and_r(1,:)
    dz_over_dr(:)=Theory%z_and_r(2,:)

    ! spline rr and dz/dr as functions of zz: then we can get r and dz/dr at any z
    call spline_real(zz,rr,num_power_redshifts,ddrr)
    call spline_real(zz,dz_over_dr,num_power_redshifts,dddz_over_dr)

    ! for each value of z in the data's selection function file, compute r and dz/dr with splint
    ! compute also the selection function eta(r) = eta(z) dz/dr normalized to one
    ! (inspired from Numerical Recipies's splint)
    nrmax=nzmax
    do nr=1,nrmax
        klo=1
        khi=num_power_redshifts
        do while (khi-klo.gt.1)
            k=(khi+klo)/2
            if(zz(k).gt.z(nr)) then
                khi=k
            else
                klo=k
            endif
        end do
        h=zz(khi)-zz(klo)
        if (h.eq.0.) then
            write(*,*)'Problem: h=0 in splint of rr(z), for nr=',nr,' z(nr)=',z(nr)
            write(*,*)'while klo=',klo,' zz(klo)=',zz(klo),' khi=',khi,' zz(khi)=',zz(khi)
            stop
        end if
        a=(zz(khi)-z(nr))/h
        b=(z(nr)-zz(klo))/h
        r(nr)=a*rr(klo)+b*rr(khi)+((a**3-a)*ddrr(klo)+(b**3-b)*ddrr(khi))*(h**2)/6.
        dz_o_dr(nr)=a*dz_over_dr(klo)+b*dz_over_dr(khi)+((a**3-a)*dddz_over_dr(klo)&
            +(b**3-b)*dddz_over_dr(khi))*(h**2)/6.
        etar_f(nr,:)=etaz_f(nr,:)*dz_o_dr(nr)/eta_norm_f(:)  
        etar_b(nr,:)=etaz_b(nr,:)*dz_o_dr(nr)/eta_norm_b(:)      
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! For each bin, compute window function g(r) = r^2 W(r) a_0/a !!!!!!!!!!!
!!!!!!!! with W(r) = 2 sum_0^r_s_max dr_s eta(r_s) (r_a-r)/r/r_s !!!!!!!!!!!!!!!
!!!!!!!! (simple trapezoidal integration ois enough) !!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    g_f(1,:)=0.
    g_b(1,:)=0.
    do nr=2,nrmax-1
        g_f(nr,:)=0.
        g_b(nr,:)=0.
        do nrs=nr+1,nrmax
            g_f(nr,:)=g_f(nr,:)+0.5*(etar_f(nrs,:)*f_k_real(r(nrs)-r(nr))/f_k_real(r(nrs)) &
                +etar_f(nrs-1,:)*f_k_real(r(nrs-1)-r(nr))/f_k_real(r(nrs-1)))*(r(nrs)-r(nrs-1))
            g_b(nr,:)=g_b(nr,:)+0.5*(etar_b(nrs,:)*f_k_real(r(nrs)-r(nr))/f_k_real(r(nrs)) &
                +etar_b(nrs-1,:)*f_k_real(r(nrs-1)-r(nr))/f_k_real(r(nrs-1)))*(r(nrs)-r(nrs-1))
        end do
    end do
    g_f(nrmax,:)=0.
    g_b(nrmax,:)=0.

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!      Fill k arrays       !!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    ! Note that we do a shift of 1 in the argument: 
    ! kk(nk) corresponds to k value (unit h/Mpc) for Theory%P_GG(nk+1,zix)

    kk(:)=exp(Theory%log_kh(2:Theory%num_k))
    nkmax=Theory%num_k-1
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! start loop over l  for computation of C_l^shear !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    do nl=1,nlmax

        ! current value of l:
        ll(nl)=1.*exp(dlnl*(nl-1.))

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!! fill array PP(r) = P(l/r,z(r)) !!!!!!!!!!!!!!
        !!! prepare it for interpolation along r  !!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        PP_GG(1)=0.
        PP_GI(1)=0.
        PP_II(1)=0.
        do zix = 2,num_power_redshifts
            kvalue=ll(nl)/f_k_real(rr(zix))/CMB%h   ! k=l/r in units h/Mpc
            if ((kvalue .le. kk(1)) .or. (kvalue .ge. kk(nkmax))) then
                PP_GG(zix)=0.
                PP_GI(zix)=0.
                PP_II(zix)=0.
            else
                PP_GG(zix)=PowerAt_zbin(Theory,Theory%P_GG,Theory%ddP_GG,kvalue,zix)
                PP_GI(zix)=PowerAt_zbin(Theory,Theory%P_GI,Theory%ddP_GI,kvalue,zix)
                PP_II(zix)=PowerAt_zbin(Theory,Theory%P_II,Theory%ddP_II,kvalue,zix)
            end if
        end do
        call spline_real(rr,PP_GG,num_power_redshifts,ddPP_GG)
        call spline_real(rr,PP_GI,num_power_redshifts,ddPP_GI)
        call spline_real(rr,PP_II,num_power_redshifts,ddPP_II)

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!! Find PP(r) = P(l/r,z(r)) for each r(nr) !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!! and infer Cl_integrand = P^{k l}_\kappa_integrand = (g_{k}(r)*g_{l}(r))*P(l/r,z(r))/r^2 !!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        
        Cl_integrand_GG(1,:)=0.
        Cl_integrand_GI(1,:)=0.
        Cl_integrand_II(1,:)=0.
        do nr=2,nrmax
        ! (inspired from Numerical Recipies's splint)
            klo=1
            khi=num_power_redshifts
            do while (khi-klo.gt.1)
                k=(khi+klo)/2
                if(rr(k).gt.r(nr)) then
                    khi=k
                else
                    klo=k
                endif
            end do
            h=rr(khi)-rr(klo)
            if (h.eq.0.) then
                write(*,*)'Problem: h=0 in splint of rr(z)'
                stop
            end if
            a=(rr(khi)-r(nr))/h
            b=(r(nr)-rr(klo))/h
            PP_GG_r=a*PP_GG(klo)+b*PP_GG(khi)+((a**3-a)*ddPP_GG(klo)+(b**3-b)*ddPP_GG(khi))*(h**2)/6.
            PP_GG_r=PP_GG_r/(CMB%h)**3  ! in units Mpc**3
            PP_GI_r=a*PP_GI(klo)+b*PP_GI(khi)+((a**3-a)*ddPP_GI(klo)+(b**3-b)*ddPP_GI(khi))*(h**2)/6.
            PP_GI_r=PP_GI_r/(CMB%h)**3  ! in units Mpc**3
            PP_II_r=a*PP_II(klo)+b*PP_II(khi)+((a**3-a)*ddPP_II(klo)+(b**3-b)*ddPP_II(khi))*(h**2)/6.
            PP_II_r=PP_II_r/(CMB%h)**3  ! in units Mpc**3
            nc=0
            do clc1=1,6 
                do clc2 = clc1,6
                    nc = nc+1
!                    if(clc2.eq.clc1) then
!                        Cl_integrand_GG(nr,nc)=g_f(nr,clc1)*g_f(nr,clc2)*PP_GG_r
!                        Cl_integrand_GI(nr,nc)=-CFHTLens_A*(g_f(nr,clc1)*etar_f(nr,clc2)+g_f(nr,clc2)*etar_f(nr,clc1))&
!                        /f_k_real(r(nr))*PP_GI_r
!                        Cl_integrand_II(nr,nc)=CFHTLens_A**2*etar_f(nr,clc1)*etar_f(nr,clc2)*PP_II_r &
!                        /f_k_real(r(nr))**2
!                    else
                        Cl_integrand_GG(nr,nc)=g_f(nr,clc1)*g_b(nr,clc2)*PP_GG_r
                        Cl_integrand_GI(nr,nc)=-CFHTLens_A*(g_f(nr,clc1)*etar_b(nr,clc2)+g_b(nr,clc2)*etar_f(nr,clc1))&
                        /f_k_real(r(nr))*PP_GI_r
                        Cl_integrand_II(nr,nc)=CFHTLens_A**2*etar_f(nr,clc1)*etar_b(nr,clc2)*PP_II_r &
                        /f_k_real(r(nr))**2
!                    end if                        
                end do
            end do
            if(nc .ne. ncl) then
                write(*,*)'nc count is ',nc, 'after loop instead of ',ncl
                stop
            end if
        end do

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!! Integrate over r to get C_l^shear !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!! C_l^shear = 9/16 Omega_m0^2 H_0^4 \sum_0^rmax dr Cl_integrand   !!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        C_GG(nl,:)=0.
        C_GI(nl,:)=0.
        C_II(nl,:)=0.
        do nr=2,nrmax
            C_GG(nl,:)=C_GG(nl,:)+0.5*(Cl_integrand_GG(nr,:)+Cl_integrand_GG(nr-1,:))*(r(nr)-r(nr-1))
            C_GI(nl,:)=C_GI(nl,:)+0.5*(Cl_integrand_GI(nr,:)+Cl_integrand_GI(nr-1,:))*(r(nr)-r(nr-1))
            C_II(nl,:)=C_II(nl,:)+0.5*(Cl_integrand_II(nr,:)+Cl_integrand_II(nr-1,:))*(r(nr)-r(nr-1))
        end do
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Spline Cl(nl,nbin) along l !!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    do nbin = 1,ncl
    ! (inspired from Numerical Recipies's spline)
        ddC_GG(1,nbin)=0.
        ddC_GI(1,nbin)=0.
        ddC_II(1,nbin)=0.
        u_spline_GG(1)=0.
        u_spline_GI(1)=0.
        u_spline_II(1)=0.
        do nl=2,nlmax-1
            sig_spline=(ll(nl)-ll(nl-1))/(ll(nl+1)-ll(nl-1))
            p_spline_GG=sig_spline*ddC_GG(nl-1,nbin)+2.
            ddC_GG(nl,nbin)=(sig_spline-1.)/p_spline_GG
            u_spline_GG(nl)=(6.*((C_GG(nl+1,nbin)-C_GG(nl,nbin))&
                /(ll(nl+1)-ll(nl))-(C_GG(nl,nbin)-C_GG(nl-1,nbin))&
                /(ll(nl)-ll(nl-1)))/(ll(nl+1)-ll(nl-1))&
                -sig_spline*u_spline_GG(nl-1))/p_spline_GG
                
            p_spline_GI=sig_spline*ddC_GI(nl-1,nbin)+2.
            ddC_GI(nl,nbin)=(sig_spline-1.)/p_spline_GI
            u_spline_GI(nl)=(6.*((C_GI(nl+1,nbin)-C_GI(nl,nbin))&
                /(ll(nl+1)-ll(nl))-(C_GI(nl,nbin)-C_GI(nl-1,nbin))&
                /(ll(nl)-ll(nl-1)))/(ll(nl+1)-ll(nl-1))&
                -sig_spline*u_spline_GI(nl-1))/p_spline_GI
                
            p_spline_II=sig_spline*ddC_II(nl-1,nbin)+2.
            ddC_II(nl,nbin)=(sig_spline-1.)/p_spline_II
            u_spline_II(nl)=(6.*((C_II(nl+1,nbin)-C_II(nl,nbin))&
                /(ll(nl+1)-ll(nl))-(C_II(nl,nbin)-C_II(nl-1,nbin))&
                /(ll(nl)-ll(nl-1)))/(ll(nl+1)-ll(nl-1))&
                -sig_spline*u_spline_II(nl-1))/p_spline_II  
        end do
        ddC_GG(nlmax,nbin)=0.
        ddC_GI(nlmax,nbin)=0.
        ddC_II(nlmax,nbin)=0.
        do nl=nlmax-1,1,-1
            ddC_GG(nl,nbin)=ddC_GG(nl,nbin)*ddC_GG(nl+1,nbin)+u_spline_GG(nl)
            ddC_GI(nl,nbin)=ddC_GI(nl,nbin)*ddC_GI(nl+1,nbin)+u_spline_GI(nl)
            ddC_II(nl,nbin)=ddC_II(nl,nbin)*ddC_II(nl+1,nbin)+u_spline_II(nl)
        end do
    end do

    if (Feedback >2) write(*,*)'lmax =',ll(nlmax)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!! Compute correlation function C_p,m(theta) (using module "AngularS") !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    call Cl_to_Ctheta(ll,C_GG,ddC_GG,C_GI,ddC_GI,C_II,ddC_II,&
                      theta,cp,cm,nlmax,nthetamax,ncl,xstop)
    

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! Arrange all C(theta) values in a vector, following the same order as in the data !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    do nr=1,nthetamax
      do nl=1,ncl
        d_theo(nr+10*(nl-1))=cp(nr,nl)
        d_theo(nr+5+10*(nl-1))=cm(nr,nl)
      end do    
    end do


    if (Feedback >2) write(*,*)'CFHTLens nuisnace parameter A =',CFHTLens_A
    ! add the chi2
    do k=1,2*nthetamax*ncl
        do nk=1,2*nthetamax*ncl 
            CFHTLensLnLike=CFHTLensLnLike+&
                (d_theo(k)-d_obs(k))*&
                invcov(k,nk)*&
                (d_theo(nk)-d_obs(nk))
        end do
    end do

    ! compute -lnL = chi2/2
    CFHTLensLnLike=CFHTLensLnLike/2.

    if (Feedback >=2) write(*,*)'CFHTLens lensing -LnLike=',CFHTLensLnLike

  END FUNCTION CFHTLensLnLike


end module CFHTLens

