!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Original module written by Julien Lesgourgues for using weak lensing data in CosmoMC        !
! It was written for the COSMOS 3D data by Massey et al. [astro-ph/0701480],                  !
! was used in [arXiv:0705.0533] by J. Lesgourgues, M. Viel, M.G. Haehnelt, R. Massey          !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This modified version was written by Jason Dossett.                                         !
! It allows the use to the refined Cosmos3D data by Schrabback et al [arXiv:0911.0053],       !
! and was first used in [arXiv:1103.1195] by J. Dossett, J. Moldenhauer, M. Ishak             !
! to probe deviations from General Relativity.                                                !
! This version makes no assumptions about the curvature and is consistent with                !
! formulae used in [arXiv:1205.2422]                                                          !
! Last Modified 01/2014                                                                       !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

MODULE WeakLen
  use Camb, only : f_k_real
  use constants, ONLY: const_pi
  use cmbtypes
  use likelihood
  use powerspec
  use AngularS ! ancillary module converting harmonic power spectrum into angular correlation functions
  implicit none
  ! precision parameters:
  integer, parameter :: nlmax=130  ! fixes the maximum required l value in C_l(lensing) 
  real, parameter :: dlnl=0.1      ! fixes logarithmic l step for C_l(lensing)
  real, parameter :: xstop=200.    ! fixes maximum value of x =  l * theta in angular correlation function 
  ! note: integration of C_l J_0,4(l*theta) over l will stop 
  ! at l = min( ll(nlmax),  xstop/theta )

  ! parameters describing Schrabback's refined Cosmos3D data set:
  
  integer, parameter :: nbinmax=6  ! Number of redshift bins
    
  integer, parameter :: nzmax = 50 ! Number of z values in galaxy distribution file
   
  integer, parameter :: nz6 = 500  ! Number of z values in data file for 
                                     ! the estimiated galaxy distribution of the sixth bin
  
  integer, parameter :: nthetamax=5 ! number of angles theta
  integer, parameter :: ncl = 16    ! number of Cross-power spectra to compute
   
  logical :: Initialize_WeakLen = .false.
!Data Arrays
  real z(nzmax),etaz(nzmax,nbinmax),eta_norm(nbinmax) ! data's selection functions
  real z6(nz6),etaz6(nz6)  ! Allows for computation of 6th bin's galaxy distribution with 
                           ! Nuissance parameter described in [arXiv:0911.0053]
  real(mcp) theta(nthetamax) ! data array of theta values
  !Arrays for calculating -ln(Like)
  real d_theo(2*nthetamax*ncl)   ! Theoretical shear cross-correlations, data arranged like a vector
  real d_obs(2*nthetamax*ncl)  ! Observed shear cross-correlations, data arranged like a vector
  real invcov(2*nthetamax*ncl,2*nthetamax*ncl)  !inverse Covariance Matrix

!Below for weighting of cross-correlations as described in [arXiv:0911.0053] and []  
  integer, parameter :: nthetastep = 100 ! for number of integration steps 
  real(mcp) dtheta(nthetamax) ! dtheta for each theta bin given number of integration steps
  real(mcp) pownorm(nthetamax) !for each theata bin Int_\thetabinmin^thetabinmax  N(\theta) d\theta
  
  !Allocatable Arrays
  !Dimension Theory%num_k-1
  real, allocatable :: kk(:)
  !Dimension num_power_redshifts
  real, allocatable :: PP(:), ddPP(:)   ! P_GG(k=l/r, z=z(r))
  !Dimension Theory%num_k-1
  real, allocatable :: u_spline(:)
  !Dimension num_power_redshifts
  real, allocatable :: zz(:),rr(:),dz_over_dr(:)
  real, allocatable :: ddrr(:),dddz_over_dr(:)
  
  type, extends(CosmologyLikelihood) :: WeakLenLikelihood
    contains
    procedure :: LogLike => WeakLenLnLike
  end type WeakLenLikelihood
  
  
CONTAINS
  
  subroutine WeakLenLikelihood_Add(LikeList, Ini)
    use IniFile
    use settings
    class(LikelihoodList) :: LikeList
    Type(TIniFile) :: ini
    Type(WeakLenLikelihood), pointer :: like
    
    use_WeakLen = Ini_Read_Logical_File(Ini, 'use_WeakLen',.false.)
    if (.not. use_WeakLen) return
    
    allocate(like)
    Like%LikelihoodType = 'WeakLen'
    Like%name='SchrLens'
    like%needs_powerspectra = .true.
    like%needs_nonlinear_pk = .true.
    like%max_z = 5
    like%num_z = 20
    like%kmax = 100
    call Like%loadParamNames(trim(DataDir)//'WeakLenSchr.paramnames')
    call LikeList%Add(like)
    
    call WeakLenInit
    
  end subroutine WeakLenLikelihood_Add
  
  
  subroutine WeakLenInit
    real b1, b2  !Dummies for reading in correlation vector
    !JD for integration of pownorm
    real(mcp) thetamax
    real(mcp) normint(4,nthetamax) ! For Simpsons rule integreation
    real(mcp) theta2(nthetamax),radtomin
    integer t,nz,k    
    character*100 name
      
      if(feedback>0)write(*,*)'Initializing refined COSMOS 3D weak lensing data'

      !Read galaxy distribution for bins 1 to 5
      name='data/COSMOS_Schrabback2010_zdist_bin1to5JD.txt'
      open (unit=10,file=name)
      do nz=1,nzmax
          read(10,*)z(nz), etaz(nz,:)
      end do
      close(10)

      !Read galaxy distribution for bin 6
      name='data/COSMOS_Schrabback2010_zdist_bin6_fit.txt'
      open (unit=10,file=name)
      do nz=1,nz6
          read(10,*)z6(nz),etaz6(nz)
      end do
      close(10)

      ! Read Thetas
      name='data/COSMOS_Schrabback2010_thetasJD.txt'
      open (unit=10,file=name)
      do k=1,nthetamax
          read(10,*)theta(k),thetamax
          theta(k)=theta(k)/60._mcp*const_pi/180._mcp
          thetamax=thetamax/60._mcp*const_pi/180._mcp
          dtheta(k)=(thetamax-theta(k))/nthetastep   !Set dtheta for each theta bin
      end do
      close(10)
      
      !read data vector
      name='data/COSMOS_Schrabback2010_correlation_vectorJD.txt'
      open (unit=10,file=name)
      read(10,*)name
      do k=1,2*nthetamax*ncl
          read(10,*)b1,b2,d_obs(k)
      end do
      close(10)


      ! Read inverse covariance matrix
      name='data/COSMOS_Schrabback2010_invcovJD.txt'
      open (unit=10,file=name)
      do k=1,2*nthetamax*ncl
          read(10,*)invcov(k,:)
      end do
      close(10)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! integrate eta(z) over z (in view of normalizing it to one) !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      eta_norm(:)=0.
      do nz=2,nzmax
          eta_norm(:)=eta_norm(:)+0.5*(etaz(nz-1,:)+etaz(nz,:))*(z(nz)-z(nz-1))
      end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Compute Bessel interpolation array !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      call GetBessels(xstop*1.d0) ! subroutine of the "Angular" module

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!! Perform normalization integral                 !!!!!
!!!! \int_{thetaxmin}^{thetaxmax} N(\theta) d\theta !!!!!
!!!! Via Simpson's 3/8's rule                       !!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      radtomin = 60._mcp*180._mcp/const_pi
      t=1
      pownorm(:) = 0._mcp
      do k=0,nthetastep-1
          theta2(:)=theta(:)+k*dtheta(:)  !vector of thetas for each step
          !fill in vector of steps
          normint(t,:)= radtomin*theta2(:)*(0.0004664+radtomin*theta2(:)&
                  *(0.0044118-8.90878e-5*radtomin*theta2(:)))
          if(t==4)then
              pownorm(:)= pownorm(:)&
                  +(3._mcp/8._mcp*(normint(1,:)+normint(4,:))&
                  +9._mcp/8._mcp*(normint(2,:)+normint(3,:)))*dtheta(:)
              normint(1,:)=normint(4,:)
              t=2
          else
              t=t+1
          end if
      end do
  end subroutine WeakLenInit  
  
  subroutine InitLens(Theory) 
  Class(TheoryPredictions) Theory
    if(allocated(PP))deallocate(PP)
    if(allocated(ddPP))deallocate(ddPP)
    if(allocated(u_spline))deallocate(u_spline)
    if(allocated(zz))deallocate(zz)
    if(allocated(rr))deallocate(rr)
    if(allocated(dz_over_dr))deallocate(dz_over_dr)
    if(allocated(ddrr))deallocate(ddrr)
    if(allocated(dddz_over_dr))deallocate(dddz_over_dr)
    if(allocated(kk))deallocate(kk)
    allocate(PP(num_power_redshifts))
    allocate(ddPP(num_power_redshifts))
    allocate(zz(num_power_redshifts))
    allocate(rr(num_power_redshifts))
    allocate(dz_over_dr(num_power_redshifts))
    allocate(ddrr(num_power_redshifts))
    allocate(dddz_over_dr(num_power_redshifts))
    allocate(kk(Theory%num_k-1))
    allocate(u_spline(Theory%num_k-1))
  end subroutine InitLens   

  function WeakLenLnLike(like,CMB,Theory,DataParams)
    Class(WeakLenLikelihood) :: like
    Class(CMBParams) CMB
    Class(TheoryPredictions) Theory
    real(mcp) DataParams(:)
    real(mcp) :: WeakLenLnLike
    
    integer nl,nc,clc1,clc2
    real ll(nlmax)
    integer nkmax,nk
    integer zix
    ! for spline:
    real sig_spline,p_spline
    ! for splint:
    integer k,khi,klo
    real a,b,h
    ! for eta(z):
    integer nz,nbin,nr,nrmax,nrs
    real r(nzmax),dz_o_dr(nzmax),etar(nzmax,nbinmax)
    real g(nzmax,nbinmax)
    ! For C_l
    real(mcp) kvalue
    real PP_r
    real Cl_integrand(nzmax,ncl)
    real C(nlmax,ncl),ddC(nlmax,ncl)
    !JD for getting new values for bin 6 with nuissance parameter involved.
    real z62(nz6),ddetaz6(nz6)
    !For C(theta)
    real Cp(nthetamax,ncl),Cm(nthetamax,ncl) !cross-correlation arrays for function Cl to Ctheta
    real WL_fz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!! Call function to allocate arrays !!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    call InitLens(Theory)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!! Various steps for Likelihood computation start here !!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! JD inclusion of Schrabback nuissance parameter !!!!!!!!
!!!!!!!! this parameter rescales the galaxy redshift !!!!!!!!!!!  
!!!!!!!! distrobution of bin 6 !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    WL_fz=DataParams(1)
    
    z62 = z6/WL_fz
    !JD prepare for cubic splin interpolation interpolation
    call spline_real(z62,etaz6,nz6,ddetaz6)
    eta_norm(6) = 0.
    
    ! for each value of z in the data's selection function file, use adjusted data from bin 6,
    ! and get a data point in bin 6 at the correct z
    ! (inspired from Numerical Recipies's splint)
    nrmax=nzmax
    do nr=1,nzmax
        klo=1
        khi=nz6
        do while (khi-klo.gt.1)
            k=(khi+klo)/2
            if(z62(k).gt.z(nr)) then
                khi=k
            else
                klo=k
            endif
        end do
        h=z62(khi)-z62(klo)
        if (h.eq.0.) then
            write(*,*)'Problem: h=0 in splint of etaz6, for nr=',nr,' z(nr)=',z(nr)
            write(*,*)'while klo=',klo,' z62(klo)=',z62(klo),' khi=',khi,' z62(khi)=',z62(khi)
            stop
        end if
        a=(z62(khi)-z(nr))/h
        b=(z(nr)-z62(klo))/h
        etaz(nr,6)=a*etaz6(klo)+b*etaz6(khi)+((a**3-a)*ddetaz6(klo)&
            +(b**3-b)*ddetaz6(khi))*(h**2)/6.
        if(nr>1) eta_norm(6)=eta_norm(6)+0.5*(etaz(nr-1,6)+etaz(nr,6))*(z(nr)-z(nr-1))
    end do


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Convert eta(z) to eta(r) !!!!!!!!!!!!!!!!!!!!
!!! (first unnormalized, second normalized to one) !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    zz(:)=Theory%redshifts(:)  ! note that z (and r, ...) must be rearranged in growing order for splint
    ! read array Theory%z_and_r(2,zix) = list of corresponding comoving distances rr(zix)
    ! read array Theory%z_and_r(3,zix) = list of corresponding dz/dr (= Hubble parameter)    
    rr(:)=Theory%z_and_r(1,:)
    dz_over_dr(:)=Theory%z_and_r(2,:)


    ! spline rr and dz/dr as functions of zz: then we can get r and dz/dr at any z
    call spline_real(zz,rr,num_power_redshifts,ddrr)
    call spline_real(zz,dz_over_dr,num_power_redshifts,dddz_over_dr)

    ! for each value of z in the data's selection function file, compute r and dz/dr with splint
    ! compute also the selection function eta(r) = eta(z) dz/dr normalized to one
    ! (inspired from Numerical Recipies's splint)
    nrmax=nzmax
    do nr=1,nrmax
        klo=1
        khi=num_power_redshifts
        do while (khi-klo.gt.1)
            k=(khi+klo)/2
            if(zz(k).gt.z(nr)) then
                khi=k
            else
                klo=k
            endif
        end do
        h=zz(khi)-zz(klo)
        if (h.eq.0.) then
            write(*,*)'Problem: h=0 in splint of rr(z), for nr=',nr,' z(nr)=',z(nr)
            write(*,*)'while klo=',klo,' zz(klo)=',zz(klo),' khi=',khi,' zz(khi)=',zz(khi)
            stop
        end if
        a=(zz(khi)-z(nr))/h
        b=(z(nr)-zz(klo))/h
        r(nr)=a*rr(klo)+b*rr(khi)+((a**3-a)*ddrr(klo)+(b**3-b)*ddrr(khi))*(h**2)/6.
        dz_o_dr(nr)=a*dz_over_dr(klo)+b*dz_over_dr(khi)+((a**3-a)*dddz_over_dr(klo)&
            +(b**3-b)*dddz_over_dr(khi))*(h**2)/6.
        etar(nr,:)=etaz(nr,:)*dz_o_dr(nr)/eta_norm(:)      
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! For each bin, compute window function g(r) = r^2 W(r) a_0/a !!!!!!!!!!!
!!!!!!!! with W(r) = 2 sum_0^r_s_max dr_s eta(r_s) (r_a-r)/r/r_s !!!!!!!!!!!!!!!
!!!!!!!! (simple trapezoidal integration ois enough) !!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    g(1,:)=0.
    do nr=2,nrmax-1
        g(nr,:)=0.
        do nrs=nr+1,nrmax
            g(nr,:)=g(nr,:)+0.5*(etar(nrs,:)*f_k_real(r(nrs)-r(nr))/f_k_real(r(nrs)) &
                +etar(nrs-1,:)*f_k_real(r(nrs-1)-r(nr))/f_k_real(r(nrs-1)))*(r(nrs)-r(nrs-1))
        end do
    end do
    g(nrmax,:)=0.

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!      Fill k arrays       !!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    ! Note that we do a shift of 1 in the argument: 
    ! kk(nk) corresponds to k value (unit h/Mpc) for Theory%P_GG(nk+1,zix)

    kk(:)=exp(Theory%log_kh(2:Theory%num_k))
    nkmax=Theory%num_k-1
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! start loop over l  for computation of C_l^shear !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    do nl=1,nlmax

        ! current value of l:
        ll(nl)=1.*exp(dlnl*(nl-1.))

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!! fill array PP(r) = P(l/r,z(r)) !!!!!!!!!!!!!!
        !!! prepare it for interpolation along r  !!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        PP(1)=0.
        do zix = 2,num_power_redshifts
            kvalue=ll(nl)/f_k_real(rr(zix))/CMB%h   ! k=l/r in units h/Mpc
            if ((kvalue .le. kk(1)) .or. (kvalue .ge. kk(nkmax))) then
                PP(zix)=0.
            else
                PP(zix)=PowerAt_zbin(Theory,Theory%P_GG,Theory%ddP_GG,kvalue,zix)
            end if
        end do
        call spline_real(rr,PP,num_power_redshifts,ddPP)

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!! Find PP(r) = P(l/r,z(r)) for each r(nr) !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!! and infer Cl_integrand = P^{k l}_\kappa_integrand = (g_{k}(r)*g_{l}(r))*P(l/r,z(r))/r^2 !!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        Cl_integrand(1,:)=0.
        do nr=2,nrmax
        ! (inspired from Numerical Recipies's splint)
            klo=1
            khi=num_power_redshifts
            do while (khi-klo.gt.1)
                k=(khi+klo)/2
                if(rr(k).gt.r(nr)) then
                    khi=k
                else
                    klo=k
                endif
            end do
            h=rr(khi)-rr(klo)
            if (h.eq.0.) then
                write(*,*)'Problem: h=0 in splint of rr(z)'
                stop
            end if
            a=(rr(khi)-r(nr))/h
            b=(r(nr)-rr(klo))/h
            PP_r=a*PP(klo)+b*PP(khi)+((a**3-a)*ddPP(klo)+(b**3-b)*ddPP(khi))*(h**2)/6.
            PP_r=PP_r/(CMB%h)**3  ! in units Mpc**3
            nc =1
            do clc1=1,5 
                do clc2 = clc1+1,6
                    Cl_integrand(nr,nc)=g(nr,clc1)*g(nr,clc2)*PP_r
                    nc = nc+1
                end do
            end do
            if(nc .ne. ncl) then
                write(*,*)'nc count is ',nc, 'after loop instead of ',ncl
                stop
            end if
            Cl_integrand(nr,nc)=g(nr,6)**2*PP_r
        end do

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!! Integrate over r to get C_l^shear !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!! C_l^shear = 9/16 Omega_m0^2 H_0^4 \sum_0^rmax dr Cl_integrand   !!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        C(nl,:)=0.
        do nr=2,nrmax
            C(nl,:)=C(nl,:)+0.5*(Cl_integrand(nr,:)+Cl_integrand(nr-1,:))*(r(nr)-r(nr-1)) 
        end do
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Spline Cl(nl,nbin) along l !!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    do nbin = 1,ncl
    ! (inspired from Numerical Recipies's spline)
        ddC(1,nbin)=0.
        u_spline(1)=0.
        do nl=2,nlmax-1
            sig_spline=(ll(nl)-ll(nl-1))/(ll(nl+1)-ll(nl-1))
            p_spline=sig_spline*ddC(nl-1,nbin)+2.
            ddC(nl,nbin)=(sig_spline-1.)/p_spline
            u_spline(nl)=(6.*((C(nl+1,nbin)-C(nl,nbin))&
                /(ll(nl+1)-ll(nl))-(C(nl,nbin)-C(nl-1,nbin))&
                /(ll(nl)-ll(nl-1)))/(ll(nl+1)-ll(nl-1))&
                -sig_spline*u_spline(nl-1))/p_spline
        end do
        ddC(nlmax,nbin)=0.
        do nl=nlmax-1,1,-1
            ddC(nl,nbin)=ddC(nl,nbin)*ddC(nl+1,nbin)+u_spline(nl)
        end do
    end do

    if (Feedback >2) write(*,*)'lmax =',ll(nlmax)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!! Compute correlation function C_p,m(theta) (using module "AngularS") !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    call Cl_to_Ctheta(ll,C,ddC,theta,dtheta,pownorm,cp,cm,nlmax,nthetamax,&
        nthetastep,ncl,xstop)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! Arrange all C(theta) values in a vector, following the same order as in the data !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    nl=0
    do nr=1,2
        do nbin=1,ncl
            do k=1,nthetamax
                nl=nl+1
                if (nr==1) then
                    d_theo(nl)=cp(k,nbin)
                else
                    d_theo(nl)=cm(k,nbin)
                end if
            end do
        end do
    end do


    ! compute chi2=-2lnL corresponding to priors on fz (modulo a constant)    
    WeakLenLnLike=((WL_fz-1)/0.1)**2
    if (Feedback >2) write(*,*)'WL nuisnace parameters f_z =',WL_fz
    ! add the chi2
    do k=1,2*nthetamax*ncl
        do nk=1,2*nthetamax*ncl 
            WeakLenLnLike=WeakLenLnLike+&
                (d_theo(k)-d_obs(k))*&
                invcov(k,nk)*&
                (d_theo(nk)-d_obs(nk))
        end do
    end do

    ! compute -lnL = chi2/2
    WeakLenLnLike=WeakLenLnLike/2.

    if (Feedback >=2) write(*,*)'Weak lensing -LnLike=',WeakLenLnLike

  END FUNCTION WeakLenLnLike


end module WeakLen

